// -*- mode: js2; indent-tabs-mode: nil; js2-basic-offset: 4 -*-
/*

    Copyright (C) 2012-2017  Dieter Schärf <ds@dschaerf.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.

 */

const Lang = imports.lang;

const Gtk = imports.gi.Gtk;
const Gvc = imports.gi.Gvc;
const St = imports.gi.St;

const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

function _getVolume() {
    if (Convenience.checkGSV(3, 4)) {
        try { return Main.panel._statusArea.volume; } catch(e) { return null; }
    } else if (Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
        try { return Main.panel.statusArea.volume; } catch(e) { return null; }
    } else {
        try { return Main.panel.statusArea.aggregateMenu._volume; } catch(e) { return null; }
    }
}

const GNOME_VOLUME = _getVolume();

const KEY_SHOW_STATUSAREA = 'show-statusarea';
const KEY_SHOW_VOLUME = 'show-volume';

const PANELMENU_BUTTON_DEFAULT = 0;
const PANELMENU_BUTTON_SHOW = 1;
const PANELMENU_BUTTON_NONE = 2;

const STATUSAREA_DEFAULT = 0;
const STATUSAREA_ADVANCED = 1;
const STATUSAREA_ALTERNATE = 2;

const VOLUME_ICON_HIGHT = {
    _output: 'audio-volume-high-symbolic',
    _input:  'microphone-sensitivity-high-symbolic'
};

const VOLUME_ICON_LOW = {
    _output: 'audio-volume-low-symbolic',
    _input:  'microphone-sensitivity-low-symbolic'
};

const VOLUME_ICON_MEDIUM = {
    _output: 'audio-volume-medium-symbolic',
    _input:  'microphone-sensitivity-medium-symbolic'
};

const VOLUME_ICON_MUTED = {
    _output: 'audio-volume-muted-symbolic',
    _input:  'microphone-sensitivity-muted-symbolic'
};

const PopupVolumeSwitchMenuItem = new Lang.Class({
    Name: 'PopupVolumeSwitchMenuItem',
    Extends: PopupMenu.PopupSwitchMenuItem,

    _init: function(control, id) {
        this._control = control;
        this._stream = this._control.lookup_stream_id(id);
        this.parent(this._stream.description, !this._stream.is_muted);
        this._connectStream = this._stream.connect('notify::is-muted', Lang.bind(this, function() {
            this.setToggleState(!this._stream.is_muted);
        }));
        this.connect('toggled', Lang.bind(this, function() {
            this._stream.change_is_muted(!this._stream.is_muted);
            if (this._switch.state) {
                this._stream.volume = this._control.get_vol_max_norm();
                this._stream.push_volume();
            }
        }));
        this.actor.connect('destroy', Lang.bind(this, function() {
            this._stream.disconnect(this._connectStream);
        }));
    },

    getIcon: function() {
        if (this._stream.port) {
            if (this._stream.port.indexOf('hdmi') >= 0) return 'video-display-symbolic';
            if (this._stream.port.indexOf('headphone') >= 0) return 'audio-headphones-symbolic';
            if (this._stream.port.indexOf('mic') >= 0) return 'audio-input-microphone-symbolic';
        }
        if (this._stream.get_gicon) {
            let iconName = this._stream.get_gicon().get_names();
            for (let i = 0; i < iconName.length; i++) {
                if (Gtk.IconTheme.get_default().has_icon(iconName[i])) return iconName[i]+'-symbolic';
            }
        }
        return 'audio-card-symbolic';
    },

    setSelected: function(value) {
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            this.setShowDot(value);
        } else {
            this.setOrnament(value ? PopupMenu.Ornament.CHECK : PopupMenu.Ornament.NONE);
        }
    }
});

const VolumeDevices = new Lang.Class({
    Name: 'VolumeDevices',
    Extends: Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8) ? PopupMenu.PopupMenuSection : PopupMenu.PopupSubMenuMenuItem,

    _init: function(control) {
        Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8) ? this.parent() : this.parent('...', true);
        this._control = control;
    },

    setSelectedItem: function(item) {
        if (this.menu) {
            this.label.set_text(item.label.get_text());
            this.icon.icon_name = item.getIcon();
        }
        if (this['selectedItem']) this.selectedItem.setSelected(false);
        this.selectedItem = item;
        this.selectedItem.setSelected(true);
    },

    addVolumeMenuItem: function(stream, getDefaultStream, setDefaultStream) {
        let item = new PopupVolumeSwitchMenuItem(this._control, stream.id);
        (this.menu || this).addMenuItem(item);
        if (stream.id == item._control[getDefaultStream]().id) this.setSelectedItem(item);
        item.connect('activate', Lang.bind(this, function(actor, event) {
            if (item._stream.id != item._control[getDefaultStream]().id) {
                item._control[setDefaultStream](item._stream);
                item._stream.change_is_muted(false);
                item.setToggleState(true);
                this.setSelectedItem(item);
                if (this.menu) {
                    if (this.menu.isOpen) this.menu.close(0);
                    if (!item._getTopMenu().isOpen) item._getTopMenu().open(-1);
                } else {
                    if (!GNOME_VOLUME.menu.isOpen) GNOME_VOLUME.menu.open(-1);
                }
            }
        }));
    }
});

const VolumeIndicator = new Lang.Class({
    Name: 'VolumeIndicator',

    _init: function() {
        this._control = GNOME_VOLUME._control;
        this._outputDevices = new VolumeDevices(this._control);
        this._inputDevices = new VolumeDevices(this._control);
        if (Convenience.checkGSV(3, 4) || Convenience.checkGSV(3, 6) || Convenience.checkGSV(3, 8)) {
            this._inputIndicator = new St.Icon({ icon_name: 'microphone-sensitivity-muted-symbolic', style_class: 'system-status-icon' });
            if (Convenience.checkGSV(3, 4)) {
                this._outputIndicator = GNOME_VOLUME.actor.get_children()[0];
                GNOME_VOLUME.actor.remove_actor(this._outputIndicator);
                this._hbox = new St.BoxLayout({ style_class: 'paneltools-panel-status-button-box' });
                this._hbox.add(this._outputIndicator);
                this._hbox.add(this._inputIndicator);
                GNOME_VOLUME.actor.add_actor(this._hbox);
                GNOME_VOLUME.menu.addMenuItem(this._inputDevices, 5);
                GNOME_VOLUME.menu.addMenuItem(this._outputDevices, 2);
            } else if (Convenience.checkGSV(3, 6)) {
                this._outputIndicator = GNOME_VOLUME.mainIcon;
                GNOME_VOLUME._box.add_actor(this._inputIndicator);
                GNOME_VOLUME._volumeMenu.addMenuItem(this._inputDevices, 5);
                GNOME_VOLUME._volumeMenu.addMenuItem(this._outputDevices, 2);
            } else if (Convenience.checkGSV(3, 8)) {
                this._outputIndicator = GNOME_VOLUME.mainIcon;
                GNOME_VOLUME._box.add_actor(this._inputIndicator);
                GNOME_VOLUME._volumeMenu.actor.remove_actor(GNOME_VOLUME._volumeMenu._input.item.actor);
                GNOME_VOLUME._volumeMenu.addMenuItem(GNOME_VOLUME._volumeMenu._input.item);
                GNOME_VOLUME._volumeMenu._input.item.addMenuItem(this._inputDevices);
                GNOME_VOLUME._volumeMenu._output.item.addMenuItem(this._outputDevices);
            }
        } else {
            this._outputIndicator = GNOME_VOLUME._primaryIndicator;
            this._inputIndicator = GNOME_VOLUME._addIndicator();
            GNOME_VOLUME._volumeMenu.addMenuItem(this._inputDevices, 2);
            GNOME_VOLUME._volumeMenu.addMenuItem(this._outputDevices, 1);
        }
        this._settings = Convenience.getSettings(Convenience.CURRENT_SETTINGS_SCHEMA);
        this._connectSettings = [
            this._settings.connect('changed::'+KEY_SHOW_STATUSAREA, Lang.bind(this, this._onShowStatusArea)),
            this._settings.connect('changed::'+KEY_SHOW_VOLUME, Lang.bind(this, this._onShowVolume))
        ];
        this._connectControl = [
            this._control.connect('state-changed', Lang.bind(this, this._onControlStateChanged)),
            this._control.connect('default-sink-changed', Lang.bind(this, this._readOutput)),
            this._control.connect('default-source-changed', Lang.bind(this, this._readInput)),
            this._control.connect('stream-added', Lang.bind(this, this._maybeShowInput)),
            this._control.connect('stream-removed', Lang.bind(this, this._maybeShowInput))
        ];
        if (Convenience.checkGSV(3, 8)) {
            this._connectVolumeMenu = [
                GNOME_VOLUME._volumeMenu.connect('headphones-changed', Lang.bind(this, this._onShowStatusArea)),
                GNOME_VOLUME._volumeMenu.connect('icon-changed', Lang.bind(this, this._onShowVolume))
            ];
        } else if (!Convenience.checkGSV(3, 4)) {
            this._connectVolumeMenu = GNOME_VOLUME._volumeMenu.connect('icon-changed', Lang.bind(this, this._onShowVolume));
        }
        this._onControlStateChanged();
    },

    _getIcon: function(property) {
        if (!this[property] || this[property].is_muted || this[property].volume <= 0) {
            return VOLUME_ICON_MUTED[property];
        } else {
            let n = Math.floor(3 * this[property].volume / this._control.get_vol_max_norm()) + 1;
            if (n < 2) return VOLUME_ICON_LOW[property];
            if (n >= 3) return VOLUME_ICON_HIGHT[property];
            return VOLUME_ICON_MEDIUM[property];
        }
    },

    _maybeShowInput: function() {
        let showInput = false;
        let recordingApps = this._control.get_source_outputs();
        if (this._input && recordingApps) {
            for (let i = 0; i < recordingApps.length; i++) {
                let outputStream = recordingApps[i];
                let id = outputStream.get_application_id();
                if (!id || (id != 'org.gnome.VolumeControl' && id != 'org.PulseAudio.pavucontrol')) {
                    showInput = true;
                    break;
                }
            }
        }
        this._showInput = showInput;
        this._updateVisibility(null, null);
    },

    _onControlStateChanged: function() {
        this._showOutput = false;
        this._showInput = false;
        (this._outputDevices.menu || this._outputDevices).removeAll();
        (this._inputDevices.menu || this._inputDevices).removeAll();
        if (this._control.get_state() == Gvc.MixerControlState.READY) {
            let sinks = this._control.get_sinks();
            if (sinks.length > 0) {
                sinks.sort(function(a, b) {
                    return Convenience.getSorted(a.description, b.description);
                });
                for (let i = 0; i < sinks.length; i++) {
                    this._outputDevices.addVolumeMenuItem(sinks[i], 'get_default_sink', 'set_default_sink');
                }
                this._readOutput();
            }
            let sources = this._control.get_sources();
            if (sources.length > 0) {
                sources.sort(function(a, b) {
                    return Convenience.getSorted(a.description, b.description);
                });
                for (let i = 0; i < sources.length; i++) {
                    this._inputDevices.addVolumeMenuItem(sources[i], 'get_default_source', 'set_default_source');
                }
                this._readInput();
            }
        } else {
            this._updateVisibility(null, null);
        }
    },

    _onShowStatusArea: function() {
        this._updateVisibility(this._settings.get_enum(KEY_SHOW_STATUSAREA), null);
    },

    _onShowVolume: function() {
        this._updateVisibility(null, this._settings.get_enum(KEY_SHOW_VOLUME));
    },

    _readInput: function() {
        if (this._connectInput) for (let i = 0; i < this._connectInput.length; i++) {
            this._input.disconnect(this._connectInput[i]);
        }
        this._input = this._control.get_default_source();
        if (this._input) {
            this._connectInput = [
                this._input.connect('notify::is-muted', Lang.bind(this, this._updateInputIndicator)),
                this._input.connect('notify::volume', Lang.bind(this, this._updateInputIndicator))
            ];
            let menuItems = (this._inputDevices.menu || this._inputDevices)._getMenuItems();
            for each(let item in menuItems) {
                if (item._stream.id == this._input.id) this._inputDevices.setSelectedItem(item);
            }
        }
        this._updateInputIndicator();
        this._maybeShowInput();
    },

    _readOutput: function() {
        if (this._connectOutput) this._output.disconnect(this._connectOutput);
        this._output = this._control.get_default_sink();
        if (this._output) {
            this._connectOutput = this._output.connect('notify::port', Lang.bind(this, this._updateOuputSliderIcon));
            let menuItems = (this._outputDevices.menu || this._outputDevices)._getMenuItems();
            for each(let item in menuItems) {
                if (item._stream.id == this._output.id) this._outputDevices.setSelectedItem(item);
            }
            this._showOutput = true;
        }
        this._updateOuputSliderIcon();
    },

    _updateInputIndicator: function() {
        this._inputIndicator.icon_name = this._getIcon('_input');
    },

    _updateOutputIndicator: function() {
        this._outputIndicator.icon_name = this._getIcon('_output');
    },

    _updateOuputSliderIcon: function() {
        if (!Convenience.checkGSV(3, 4) && !Convenience.checkGSV(3, 6) && !Convenience.checkGSV(3, 8)) GNOME_VOLUME._volumeMenu._output._icon.icon_name = 'audio-speakers-symbolic';
    },

    _updateVisibility: function(showStatusArea, showVolume) {
        if (showStatusArea == null) showStatusArea = this._settings.get_enum(KEY_SHOW_STATUSAREA);
        if (showVolume == null) showVolume = this._settings.get_enum(KEY_SHOW_VOLUME);
        let showOutput = this._showOutput;
        if (showVolume != PANELMENU_BUTTON_DEFAULT) showOutput = showOutput && showVolume != PANELMENU_BUTTON_NONE;
        let showInput = this._showInput;
        if (showStatusArea != STATUSAREA_DEFAULT) showInput = showInput || (this._input && showStatusArea == STATUSAREA_ALTERNATE);
        if (Convenience.checkGSV(3, 4)) {
            GNOME_VOLUME.actor.visible = showOutput;
            GNOME_VOLUME._inputTitle.actor.visible = showInput;
            GNOME_VOLUME._inputSlider.actor.visible = showInput;
        } else if (Convenience.checkGSV(3, 6)) {
            GNOME_VOLUME.actor.visible = showOutput;
            GNOME_VOLUME._volumeMenu._inputTitle.actor.visible = showInput;
            GNOME_VOLUME._volumeMenu._inputSlider.actor.visible = showInput;
        } else if (Convenience.checkGSV(3, 8)) {
            GNOME_VOLUME.actor.visible = showOutput;
            GNOME_VOLUME._headphoneIcon.visible = GNOME_VOLUME._volumeMenu.hasHeadphones && showStatusArea == STATUSAREA_DEFAULT;
            GNOME_VOLUME._volumeMenu._input._title.actor.visible = showInput;
            GNOME_VOLUME._volumeMenu._input._slider.actor.visible = showInput;
        } else if (Convenience.checkGSV(3, 9)) {
            GNOME_VOLUME.actor.visible = showOutput;
            GNOME_VOLUME._volumeMenu._input._slider.actor.visible = showInput;
        } else {
            GNOME_VOLUME.indicators.visible = showOutput;
            GNOME_VOLUME.menu.actor.visible = showOutput;
            GNOME_VOLUME._volumeMenu._input.item.actor.visible = showInput;
        }
        if (showOutput && this._outputIndicator.icon_name == null) this._updateOutputIndicator();
        this._outputIndicator.visible = showOutput;
        this._inputIndicator.visible = showOutput && showInput;
        this._outputDevices.actor.visible = showOutput && (this._outputDevices.menu || this._outputDevices).numMenuItems > 1 && showStatusArea != STATUSAREA_DEFAULT;
        this._inputDevices.actor.visible = showInput && (this._inputDevices.menu || this._inputDevices).numMenuItems > 1 && showStatusArea != STATUSAREA_DEFAULT;
    },

    destroy: function() {
        if (this._connectOutput) this._output.disconnect(this._connectOutput);
        if (this._connectInput) for (let i = 0; i < this._connectInput.length; i++) {
            this._input.disconnect(this._connectInput[i]);
        }
        if (Convenience.checkGSV(3, 8)) {
            for each(let connection in this._connectVolumeMenu) {
                GNOME_VOLUME._volumeMenu.disconnect(connection);
            }
        } else if (!Convenience.checkGSV(3, 4)) {
           GNOME_VOLUME._volumeMenu.disconnect(this._connectVolumeMenu);
        }
        for each(let connection in this._connectControl) {
            this._control.disconnect(connection);
        }
        for each(let connection in this._connectSettings) {
            this._settings.disconnect(connection);
        }
        this._updateVisibility(STATUSAREA_DEFAULT, PANELMENU_BUTTON_DEFAULT);
        if (Convenience.checkGSV(3, 4)) {
            this._hbox.remove_actor(this._outputIndicator);
            GNOME_VOLUME.actor.remove_actor(this._hbox);
            GNOME_VOLUME.actor.add_actor(this._outputIndicator);
            this._hbox.destroy();
        }
        this._inputDevices.destroy();
        this._outputDevices.destroy();
        this._inputIndicator.destroy();
        if (!Convenience.checkGSV(3, 4) && !Convenience.checkGSV(3, 6) && !Convenience.checkGSV(3, 8)) GNOME_VOLUME._volumeMenu._output._updateSliderIcon();
    }
});

function getExtension() {
    return GNOME_VOLUME ? new VolumeIndicator() : null;
}
